/****************************************************************************
*****************************************************************************

FILE NAME: list.c

PROJECT: libHLS

AUTHOR: Shantanu Tarafdar    DATE: Nov 10, 1996

ABSTRACT: Functions to manipulate lists.

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General1 Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

COPYRIGHT (C) 1996, Shantanu Tarafdar (shantanu@ece.neu.edu)
                                      (shantanu@ee.cornell.edu)

*****************************************************************************
****************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <signal.h>

#include <malloc.h>
#include <unistd.h>

#include "list.h"

/* Create a list. This must be called to initialize a list */

List *L_Create()
{
  List *l;

  /* Allocate the list memory */
  l = (List *) calloc (1, sizeof(List));
  if (l == NULL)
    L_Error("L_Create() couldn't alloc mem.");
  
  /* Initialize the list */
  l->head = NULL;
  l->tail = NULL;
  l->curr = NULL;
  l->size = 0;
  
  return l;
}

/* Insert an item in the list at the position. 
   Position = HEAD   =>   Insert at head of list
            = TAIL   =>   Insert at tail of list
            = CURR   =>   Insert at current position. Previous current
                          item gets pushed one back. 
*/

int L_Insert(List *l, void *item, int position)
{
  L_Item *li;  /* New list item */

  assert(l && item);

  /* Check if list is ok */
  if (l == NULL)
    return L_NULL;

  /* Create the new item */
  li = (L_Item *)calloc(1, sizeof(L_Item));
  if (li == NULL)
    L_Error("L_Insert() couldn't alloc mem for new item.");

  /* Link item to the list item */
  li->item = item;

  /* If list is empty then just put the new item in */
  if (l->size == 0){
    l->head = l->tail = l->curr = li;
    l->size++;
    li->prv = NULL;
    li->nxt = NULL;
    return L_OK;
  }

  /* Otherwise, you have three different situations */

  switch (position) {
    
  case L_HEAD:  
    /* Inserting item at head of list */
    l->head->prv = li;
    li->nxt = l->head;
    li->prv = NULL;
    l->head = li;
    break;
  case L_TAIL:
    /* Inserting item at tail of list */
    l->tail->nxt = li;
    li->nxt = NULL;
    li->prv = l->tail;
    l->tail = li;
    break;
  case L_CURR:
    /* Inserting item at current position */
    if (l->curr == NULL)
      L_Error("L_Insert() detected l->curr = NULL");
    li->nxt = l->curr;
    li->prv = l->curr->prv;
    l->curr->prv = li;
    if (li->prv != NULL)
      li->prv->nxt = li;
    else
      l->head = li;
    l->curr = li;
    break;
  default:
    L_Error("L_Insert detected invalid position argument.");
  }

  l->size++;
  return L_OK;
}
    
/* Removes the item specified from the list. Position arguments are the
   same as those in L_Insert */

void *L_Remove(List *l, int position)
{
  void *i = NULL;    /* Item removed from list */
  L_Item *li; /* Temporary store for list item ptr */

  /* Check if list is ok */
  if (l == NULL)
    return NULL;

  /* If list is empty return a NULL */
  if (l->size == 0)
    return NULL;

  /* If list has only one element remove it */
  if (l->size == 1){
    i = l->head->item;
    free(l->head);
    l->head = l->tail = l->curr = NULL;
    l->size = 0;
    return i;
  }

  /* Otherwise you have three options */
  switch(position){
  case L_HEAD:
    /* Remove item from head of list */
    l->head->nxt->prv = NULL;
    i = l->head->item;
    li = l->head->nxt;
    l->curr = (l->curr == l->head) ? li : l->curr;
    free(l->head);
    l->head = li;
    l->size--;
    break;
  case L_TAIL:
    /* Remove item from tail of list */
    l->tail->prv->nxt = NULL;
    i = l->tail->item;
    li = l->tail->prv;
    l->curr = (l->curr == l->tail) ? li : l->curr;
    free(l->tail);
    l->tail = li;
    l->size--;
    break;
  case L_CURR:
    /* If l->curr == NULL, then we are already at tail, so ignore */
    if (!l->curr)
      break;
    /* Remove item from current position */
    i = l->curr->item;
    if (l->curr->prv != NULL)
      l->curr->prv->nxt = l->curr->nxt;
    else
      l->head = l->curr->nxt;
    if (l->curr->nxt != NULL)
      l->curr->nxt->prv = l->curr->prv;
    else
      l->tail = l->curr->prv;
    /* li = (l->curr->nxt != NULL) ? l->curr->nxt : l->curr->prv; */
    li = l->curr->nxt;
    free(l->curr);
    l->curr = li;
    l->size--;
    break;
  default:
    L_Error("L_Remove() detected an invalid position argument.");
  }

  return i;
}

/* Search for an item in the list and removes it from the list.
   Use cmpFnc as the compare function. If cmpFnc is NULL, then
   compare based on pointers to the item. cmpFnc compares the two
   data items and returns a 1 if they are the same. cmpFnc is user
   specified. */

int L_Search_N_Remove(List *l, void *item, int (* cmpFnc) (void *a, void*b))
{
  L_Item *curr_li;
  
  /* Sanity check */
  if ((l == NULL))
    return L_NOEXIST;

  /* Preserve the current item pointer and the next item pointer. */
  curr_li = l->curr;
  
  /* Search for the item */
  if (L_Search(l, item, cmpFnc) == L_OK) {
    /* If the new current item pointer is at the old current item pointer */
    if (l->curr == curr_li) {
      L_Remove(l, L_CURR); 
    }
    /* if it is at some other place, remove the item and reset the pointer */
    else {
      L_Remove(l, L_CURR);
      l->curr = curr_li;
    }
    return L_OK;
  }
  else {
    return L_NOEXIST;
  }
  
}

/* Moves the current position in the list in a specified direction by
   a specified number of items. */

int L_Move(List *l, int dir, int no_of_items){
  int c; /* Tmp cntr */
  
  /* Check if the list is ok */
  if (l == NULL)
    return L_NULL;

  /* If the direction is L_HEAD or L_TAIL then move the pointer */
  if (dir == L_HEAD){
    l->curr = l->head;
    return L_OK;
  }
  if (dir == L_TAIL){
    l->curr = l->tail;
    return L_OK;
  }
    
  /* Set up the counter */
  c = no_of_items;

  /* Move the current ptr UP towards the head of the list */
  if (dir == L_UP){
    while((c > 0) && (l->curr->prv != NULL)){
      l->curr = l->curr->prv;
      c--;
    }
  }
  /* Move the current ptr DN towards the tail of the list */
  else if (dir == L_DN){
    while((c > 0) && (l->curr->nxt != NULL)){
      l->curr = l->curr->nxt;
      c--;
    }
  }
  else
    L_Error("L_Move() detected an invalid direction.");

  return L_OK;
}
      
/* Search for an item in the list and position the current ptr at that
   item. Use cmpFnc as the compare function. If cmpFnc is NULL, then
   compare based on pointers to the item. cmpFnc compares the two
   data items and returns a 1 if they are the same. cmpFnc is user
   specified. */

int L_Search(List *l, void *item, int (* cmpFnc) (void *a, void*b))
{
  L_Item *li; /* Tmp list item ptr store */

  /* Check if list is ok */
  if (l == NULL)
    return L_NULL;

  if (l_size(l) == 0) return L_NOEXIST;
  
  /* Start at head of list and search until cmpFnc returns a 1 or if
     cmpFnc is NULL until the current item has the same pointer as
     the one being searched for. */

  li = l->head;
  
  while (li != NULL){
    if (cmpFnc != NULL){
      if (cmpFnc(li->item, item) == 1){
	l->curr = li;
	return L_OK;
      }
      else
	li = li->nxt;
    }
    else{
      if (li->item == item){
	l->curr = li;
	return L_OK;
      }
      else
	li = li->nxt;
    }
  }

  return L_NOEXIST;
}

/* Same as L_Search() except the current item pointer is not repositioned */

int L_LookFor(List *l, void *item, int (* cmpFnc) (void *a, void*b))
{
  L_Item *_li;
  int res;

  if (l == NULL)
    return L_NOEXIST;

  _li = l->curr;
  res = L_Search(l, item, cmpFnc);
  l->curr = _li;

  return res;
}
  
/* Returns the item at the specified position in the list */

void *L_Get(List *l, int position)
{
  /* Check if the list is ok */
  if (l == NULL)
    return NULL;

  /* If list is empty then return a NULL */
  if (l->size == 0)
    return NULL;

  /* Otherwise you have three options */
  switch(position){
  case L_HEAD:
    return l->head->item;
  case L_TAIL:
    return l->tail->item;
  case L_CURR:
    return l->curr->item;
  default:
    L_Error("L_Get() detected an invalid position.");
  }

  return NULL;
}

/* Destroys the list and cleans up all the allocated memory.
   freeFnc is a function that the user specifies to tell L_Destroy()
   what to do with the items in the list. If freeFnc is NULL, then
   L_Destroy does nothing with the items and it is assumed that the
   user has been keeping track of the pointers to them. */

void L_Destroy(List *l, void (*freeFnc) (void *a))
{
  /* Check if the list is ok */
  if (l == NULL)
    return;

  /* Remove all the items */
  while (l->size != 0)
    if (freeFnc != NULL)
      freeFnc(L_Remove(l, L_HEAD));
    else
      L_Remove(l, L_HEAD);

  /* Free the list */
  free(l);

  return;
}

/* This is an internal function that displays an internal error message on
   stderr and issuing a SIGHUP to stop the process */

void L_Error(char *msg)
{
  fprintf(stderr, "Internal Error from list.c: %s\n Exiting...\n", msg);
  kill(getpid(), SIGHUP); /* Irmi: kill is not declared anyway ?! */
}

/* This is a function that executes a particular function on every element
   of the list. All the input and output parameters must be in the args data
   structure. The function being executed must be of type 
   int ExecFn(void *args) 
   The return value of ExecFn must be a 0 if the function executed correctly
   and a non-zero value if it encountered an error. L_Execute then returns
   the non-zero value to the calling program, stops executing ExecFn on the
   list and leaves the list pointer curr pointing to the list element that
   caused the error. */

int L_Execute(List *l, int (*ExecFn) (void *item, void *args), void *args)
{
  int c1;
  int retVal;
  void *item;
  L_Item *li;

  /* Check if the list is ok */
  if (l == NULL)
    return 0;

  /* Reset list pointer to the head of the list */
  L_Move(l, L_HEAD, 0);

  /* For each element in the list perform ExecFn */
  for (c1=0; c1<L_Size(l); c1++){
    /* Get the current list item */
    item = L_Get(l, L_CURR);
    /* Execute the function on the current list element */
    /* Inserted a store for l->curr in case ExecFn changes it */
    li = l->curr;
    retVal = ExecFn(item, args);
    l->curr = li;
    /* If retVal is not 0, an error occurred. So return retVal and stop */
    if (retVal != 0)
      return retVal;
    L_Move(l, L_DN, 1);
  }

  return 0;
}

/* Generates a copy of the list. Does not copy the items in the list though.
   Just regenerates the pointers in the new list. */

List *L_Copy(List *l)
{
  List *nl;   /* New list */
  L_Item *_li;  /* Place to save the current item list pointer */
  int c1;
  void *i;

  /* Check if list is ok */
  if (l == NULL)
    return NULL;

  nl = L_Create();
  
  /* Save the current item pointer */
  _li = l->curr;

  L_Move(l, L_HEAD, 0);
  for (c1=0; c1<L_Size(l); c1++){
    i = L_Get(l, L_CURR);
    L_Insert(nl, i, L_TAIL);
    L_Move(l, L_DN, 1);
  }

  /* Restore the current item pointer */
  l->curr = _li;

  return nl;
}

/* Uses merge sort to sort the list. The user MUST supply a compare function.
   The compare function gets two items as its arguments. If the first is
   "greater" than the second then the function returns a 1. If the first is
   equal to the second, it returns a 0. If the first is less than the second
   it returns a -1. Merge sort returns a list with the greatest item first */

/* 8/21/98 The return values for cmpFnc seem to be backwards.  Should
   return 1 when a is less than b and -1 when a is greater than b - VJO */

void L_Sort(List *l, int (* cmpFnc) (void *a, void *b))
{
  int n, c1;
  List *l1, *l2;

  if (!l || !cmpFnc)
    return;

  if (L_Size(l) == 1 || L_Size(l) == 0)
    return;

  n = L_Size(l)/2;

  l1 = L_Create();
  l2 = L_Create();

  L_Move(l, L_HEAD, 0);
  for (c1=0; c1<n; c1++){
    L_Insert(l1, L_Get(l, L_CURR), L_TAIL);
    L_Move(l, L_DN, 1);
  }
  for (c1=0; c1<L_Size(l)-n; c1++){
    L_Insert(l2, L_Get(l, L_CURR), L_TAIL);
    L_Move(l, L_DN, 1);
  }
 
  L_Sort(l1, cmpFnc);
  L_Sort(l2, cmpFnc);

  l1->curr = l1->head;
  l2->curr = l2->head;

  /* Empty original list prior to refilling it sorted */
  L_Move(l, L_HEAD, 0);
  while(L_Size(l) > 0){
    L_Remove(l, L_CURR);
  }

  l1->curr = l1->head;
  l2->curr = l2->head;

  for (c1=0; c1<(L_Size(l1)+L_Size(l2)); c1++){
    if (!l1->curr && !l2->curr)
      break;
    else if (!l1->curr)
      n = -1;
    else if (!l2->curr)
      n = 1;
    else
      n = cmpFnc(l1->curr->item, l2->curr->item);

    switch(n){
    case 1:
    case 0:
      L_Insert(l, l1->curr->item, L_TAIL);
      l1->curr = l1->curr->nxt;
      break;
    case -1:
      L_Insert(l, l2->curr->item, L_TAIL);
      l2->curr = l2->curr->nxt;
      break;
    default:
      L_Error("cmpFnc returned invalid value in L_Sort.");
    }   
  }

  L_Destroy(l1, NULL);
  L_Destroy(l2, NULL);

  return;
}

/* Merges two lists by performing an intersection. cmpFnc is a function
   that returns a 1 if a and b are equivalent. If cmpFnc is NULL, the 
   pointers are compared. The merged list is returned in l1. */

void L_Merge(List *l1, List *l2, int (* cmpFnc) (void *a, void *b))
{
  l_rwd(l2);
  while(!l_atend(l2)) {
    void *a = l_get(l2);
    if (L_Search(l1, a, cmpFnc) == L_NOEXIST)
      l_prepend(l1, a);
    l_nxt(l2);
  }
}

/*****************************************************************************

FUNCTION: L_Concat

AUTHOR: Shantanu Tarafdar     DATE: Mar 10, 1997

ABSTRACT: Concatenates two lists. Appends list 2 to the end of list 1. 
          Doesn't check for duplicate items or anything sophisticated
	  like that.

RETURNS: Nothing

*****************************************************************************/

void L_Concat(List *l1, List *l2)
{
  void *item;

  /* Sanity check */
  assert(l1 != NULL);
  assert(l2 != NULL);
  
  l_rwd(l2);
  while(!l_atend(l2)) {
    item = l_get(l2);
    l_append(l1, item);
    l_nxt(l2);
  }
}

/*****************************************************************************

FUNCTION: L_Randomize

AUTHOR: Shantanu Tarafdar     DATE: Jul 18, 1997

ABSTRACT: Randomizes the order of elements in a list.

RETURNS: 

*****************************************************************************/
#if 0

void L_Randomize(List *l)
{
  List *ll;
  struct timeval tv;
  unsigned long a, b, seed;
  int n, index;
  void *d;

  /* Set the seed for the random number generator */
  gettimeofday(&tv, NULL);
  /* Use the XOR of the two components of tv for the seed */
  a = (unsigned long) tv.tv_usec;
  b = (unsigned long) tv.tv_sec;
  seed = (a * b);
  srand(seed);

  /* Preserve a temporary copy of the list and empty the original one */
  ll = l_copy(l);
  l_empty(l);

  /* Until all elements in ll have been used up */
  while (!l_isempty(ll)) {
    /* Generate a random number and an index into the ll */
    n = rand();
    index = (int)((float)((float)(n)/((float)((1<<15)-1)))*
		  ((float)(l_size(ll)-1)));
    assert(index < l_size(ll));
    /* Get the element on ll at that index */
    l_rwd(ll);
    L_Move(ll, L_DN, index);
    d = l_get(ll);
    /* Append the element to l */
    l_append(l, d);
    /* Remove the element from ll */
    l_rm(ll);
  }

  /* Cleanup */
  l_free(ll);

}
#endif
/*****************************************************************************

FUNCTION: L_Reverse

AUTHOR: Shantanu Tarafdar     DATE: Oct  2, 1997

ABSTRACT: Flips a list end for end. Suggested by Valerie Ohm.

RETURNS: Nothing 

*****************************************************************************/

void L_Reverse(List *l)
{
  L_Item *li; 
  L_Item *nxt_li;

  /* If the list has size 0 or 1, there is no need for reversal */
  if ((l->size == 0) || (l->size == 1))
    return;

  /* To reverse the list, we simply have to interchange the
     nxt and prv pointers for every list item and swap the lists
     head and tail */

  li = l->head;
  nxt_li = li->nxt;
  while (nxt_li != NULL) {
    nxt_li = li->nxt;
    li->nxt = li->prv;
    li->prv = nxt_li;
    li = nxt_li;
  }
  

  li = l->head;
  l->head = l->tail;
  l->tail = li;

  /* All done */
}

/*****************************************************************************

FUNCTION: L_Intersect

AUTHOR: Shantanu Tarafdar     DATE: Oct 24, 1997

ABSTRACT: Returns a list that is the intersection of two lists. Suggested
          by Valerie Ohm.
	  
	  cmpFnc should return 1 if a == b and 0 otherwise.

RETURNS: Pointer to list containing intersection.

*****************************************************************************/

List *L_Intersect(List *l1, List *l2, int (* cmpFnc) (void *a, void *b))
{
  List *l;
  L_Item *li1 = l1->curr;
  L_Item *li2 = l2->curr;

  if (!l1 || !l2)
    return NULL;
  
  l = l_new();
  
  l1->curr = l1->head;
  while (l1->curr) {
    l2->curr = l2->head;
    while (l2->curr) {
      if (cmpFnc) {
	if (cmpFnc(l1->curr->item, l2->curr->item) == 1){
	  l_append(l, l1->curr->item);
	  break;
	}
      }
      else {
	if (l1->curr->item == l2->curr->item) {
	  l_append(l, l2->curr->item);
	  break;
	}
      }
      l2->curr = l2->curr->nxt;
    }
    l1->curr = l1->curr->nxt;
  }
    
  l1->curr = li1;
  l2->curr = li2;

  return l;

}

/* my functions (xtina) */


void free_list(list *l) {

  l_item *li;
  l_item *linx;

  if(l == NULL)
    return;
  
  
  li = l->head;

  while(li != NULL) {
    linx = li->nxt;
    li->nxt = li->prv = NULL;
    li->item = NULL;
    free(li);
    li = linx;
  }
  l->head = l->curr = l->tail = NULL;
  free(l);

  return;
}
