/****************************************************************************
*****************************************************************************

FILE NAME: list.h

PROJECT: libHLS

AUTHOR: Shantanu Tarafdar    DATE: Nov 10, 1996

ABSTRACT: Header file for the list data structure and manipulation library.

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

COPYRIGHT (C) 1996, Shantanu Tarafdar (shantanu@ece.neu.edu)
                                      (shantanu@ee.cornell.edu)

*****************************************************************************
****************************************************************************/

#ifndef _hls_list_h
#define _hls_list_h 

#define L_HEAD 1
#define L_CURR 2
#define L_TAIL 3
#define L_UP   4
#define L_DN   5

#define L_NULL -1
#define L_OK 1
#define L_NOEXIST 0

#define true 1
#define false 0

/*****************************************************************************

TYPE: L_Item

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: A list is composed of one or more L_Items which point to each other.
          They form a doubly-linked list. The first points to NULL as its
          previous L_Item and the last points to NULL as its next L_Item.
          Each L_Item stores the actual data at that point in the list under
          its item member. 

*****************************************************************************/

typedef struct _L_Item{
  void *item;            /* The actual list datum */
  struct _L_Item *nxt;   /* The next item on the list */
  struct _L_Item *prv;   /* The previous item on the list */
} L_Item;

/*****************************************************************************

TYPE: List

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: The data structure for the list. The head and tail keep track of the
          two ends of the list. The curr member keeps track of the current
          item being processed.

*****************************************************************************/

typedef struct _List{
  int size;
  L_Item *head;
  L_Item *tail;
  L_Item *curr;
} List;


/*my macros (xtina) */
#define l_item L_Item  
#define list List  
#define new_list L_Create  
#define l_insert L_Insert  
#define l_remove L_Remove  


/* Macros */

/*****************************************************************************

MACRO: l_new

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Creates a new list 

*****************************************************************************/

#define l_new L_Create

/*****************************************************************************

MACRO: l_size

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Returns the size of the list 

*****************************************************************************/

#define l_size L_Size

/*****************************************************************************

MACRO: l_isempty

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Returns a true if the list is empty and a false otherwise 

*****************************************************************************/

#define l_isempty L_IsEmpty

/*****************************************************************************

MACRO: l_rwd

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Rewinds the list pointer to the head of the list 

*****************************************************************************/

#define l_rwd(A) L_Move(A, L_HEAD, 0)

/*****************************************************************************

MACRO: l_fwd

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Moves the list pointer to the tail of the list 

*****************************************************************************/

#define l_fwd(A) L_Move(A, L_TAIL, 0)

/*****************************************************************************

MACRO: l_prepend

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Add the data to the head of the list

*****************************************************************************/

#define l_prepend(A, B) L_Insert(A, B, L_HEAD)

/*****************************************************************************

MACRO: l_append

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Adds the data to the tail of the list 

*****************************************************************************/

#define l_append(A, B) L_Insert(A, B, L_TAIL)

/*****************************************************************************

MACRO: l_ins

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Inserts the data at the current point in the list  

*****************************************************************************/

#define l_ins(A,B) L_Insert(A, B, L_CURR)

/*****************************************************************************

MACRO: l_get

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Gets the data at the current point in the list  

*****************************************************************************/

#define l_get(A) L_Get(A, L_CURR)

/*****************************************************************************

MACRO: l_first

AUTHOR: Shantanu Tarafdar     DATE: Oct 30, 1996

ABSTRACT: Gets the data at the head of the list 

*****************************************************************************/

#define l_first(A) L_Get(A, L_HEAD)

/*****************************************************************************

MACRO: l_last

AUTHOR: Shantanu Tarafdar     DATE: Oct 30, 1996

ABSTRACT: Gets the data at the tail of the list

*****************************************************************************/

#define l_last(A) L_Get(A, L_TAIL)

/*****************************************************************************

MACRO: l_nxt

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Moves the list pointer one forward in the list  

*****************************************************************************/

#define l_nxt(A) if(A && A->curr) A->curr = A->curr->nxt

/*****************************************************************************

MACRO: l_prv

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Moves the list pointer one backward in the list  

*****************************************************************************/

#define l_prv(A) if(A && A->curr) A->curr = A->curr->prv

/*****************************************************************************

MACRO: l_atend

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Returns a true if the pointer is at either end of the list and a 
          false if it is not.

*****************************************************************************/

#define l_atend(A) ((A == NULL) ? true : ((A->curr == NULL) ? true : false))

/*****************************************************************************

MACRO: l_find

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Looks for a item in the list based on its pointer alone.

*****************************************************************************/

#define l_find(A, B) L_Search(A, B, NULL)

/*****************************************************************************

MACRO: l_lookfor

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Looks for a item in the list based on its pointer alone. But 
          doesn't reposition the current item pointer.

*****************************************************************************/

#define l_lookfor(A, B) L_LookFor(A, B, NULL)

/*****************************************************************************

MACRO: l_rm

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Removes the item in the list that the list pointer is currently at

*****************************************************************************/

#define l_rm(A) L_Remove(A, L_CURR)

/*****************************************************************************

MACRO: l_behead

AUTHOR: Shantanu Tarafdar     DATE: Oct 30, 1997

ABSTRACT: Removes the first item in the list. Suggested by Valerie Ohm.

*****************************************************************************/

#define l_behead(A) L_Remove(A, L_HEAD)

/*****************************************************************************

MACRO: l_betail

AUTHOR: Shantanu Tarafdar     DATE: Oct 30, 1997

ABSTRACT: Removes the last item in the list. Suggested by Valerie Ohm.

*****************************************************************************/

#define l_betail(A) L_Remove(A, L_TAIL)

/*****************************************************************************

FUNCTION: l_del

AUTHOR: Shantanu Tarafdar     DATE: Oct  4, 1996

ABSTRACT: Searches for the item based on its pointer and removes it from the
          list.

*****************************************************************************/

#define l_del(A, B) L_Search_N_Remove(A, B, NULL)

/*****************************************************************************

MACRO: l_free

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Frees up the memory for the list without attempting to free the 
          data that make up the list. The user is responsible for that. 

*****************************************************************************/

#define l_free(A) L_Destroy(A, NULL)

/*****************************************************************************

MACRO: l_copy

AUTHOR: Shantanu Tarafdar     DATE: Sep 25, 1996

ABSTRACT: Makes a copy of the list

*****************************************************************************/

#define l_copy(A) L_Copy(A)

/*****************************************************************************

MACRO: l_sforeach

AUTHOR: Shantanu Tarafdar     DATE: Oct  1, 1996

ABSTRACT: This is used to iterate through every item on the list. A is the
          list and B is a void pointer. B gets the pointer to the current
          iteration of the for loop.

	  Do not return out of the loop body. This will corrupt the current
          item pointer of the list. Instead use break and continue.

	  Eg. l_sforeach (l, i) {
               ...
               <your code goes here>
               ...
               } l_endfor;
	  
*****************************************************************************/

#define l_sforeach(A, B) \
{ \
  List *_l = l_copy(A); \
  int first_iter = 1; \
  int fast = 0; \
  \
  l_rwd(_l); \
  while (1) { \
    if (first_iter) \
      first_iter=0; \
    else \
      l_nxt(_l); \
    if (l_atend(_l)) \
      break; \
    B = l_get(_l); \
    { 

/*****************************************************************************

MACRO: l_fforeach

AUTHOR: Shantanu Tarafdar     DATE: May  2, 1997

ABSTRACT: A fast version of l_sforeach. This has the drawback of not allowing
          the user to remove items from the list while in the body of the
	  loop.

	  On request of Steve Strickland. Happy birthday, Steve!

*****************************************************************************/

#define l_fforeach(A, B) \
{ \
  List *_l = A; \
  L_Item *_li = NULL; \
  int first_iter = 1; \
  int fast = 1; \
  \
  while (1) { \
    if (first_iter) { \
      _li = _l->head; \
      first_iter=0; \
    } \
    else { \
      _li = _li->nxt; \
    } \
    if (!_li) \
      break; \
    B = _li->item; \
    { 

/*****************************************************************************

MACRO: l_endfor

AUTHOR: Shantanu Tarafdar     DATE: Oct  1, 1996

ABSTRACT: This statement must end a l_sforeach loop.

*****************************************************************************/

#define l_endfor \
    } \
  }\
  if (!fast) l_free(_l);\
}

/*****************************************************************************

MACRO: l_merge

AUTHOR: Shantanu Tarafdar     DATE: Sep 26, 1996

ABSTRACT: Creates a "union" of two lists using the pointers to data as the
          basis for comparison.

*****************************************************************************/

#define l_merge(A, B) L_Merge(A, B, NULL)

/*****************************************************************************

MACRO: l_sort

AUTHOR: Shantanu Tarafdar     DATE: Sep 26, 1996

ABSTRACT: Wrapper for L_Sort 

*****************************************************************************/

#define l_sort L_Sort

/*****************************************************************************

MACRO: l_empty

AUTHOR: Shantanu Tarafdar     DATE: Jun  3, 1997

ABSTRACT: Empties a list 

*****************************************************************************/

#define l_empty(A) { l_rwd(A); while(!l_isempty(A)) l_rm(A); }

/*****************************************************************************

MACRO: l_concat

AUTHOR: Shantanu Tarafdar     DATE: Mar 10, 1997

ABSTRACT: Wrapper for L_Concat

*****************************************************************************/

#define l_concat L_Concat

/*****************************************************************************

MACRO: l_randomize

AUTHOR: Shantanu Tarafdar     DATE: Aug  1, 1997

ABSTRACT: Wrapper for L_Randomize

*****************************************************************************/

#define l_randomize L_Randomize

/*****************************************************************************

MACRO: L_IsEmpty

AUTHOR: Shantanu Tarafdar     DATE: Sep 26, 1996

ABSTRACT: Returns true if the list is empty and a false otherwise

*****************************************************************************/

#define L_IsEmpty(A) ((A == NULL) ? L_NULL : ((A->size == false) ? true : false))

/*****************************************************************************

MACRO: L_Size

AUTHOR: Shantanu Tarafdar     DATE: Sep 26, 1996

ABSTRACT: Returns the size (number of elements/data) of the list

*****************************************************************************/

#define L_Size(A) ((A == NULL) ? L_NULL : (A)->size)

/*****************************************************************************

MACRO: l_intersect

AUTHOR: Shantanu Tarafdar     DATE: Oct 24, 1997

ABSTRACT: Wrapper around L_Intersect

*****************************************************************************/

#define l_intersect(A, B) L_Intersect(A, B, NULL)

/*****************************************************************************

FUNCTION: l_reverse

AUTHOR: Valerie Ohm     DATE: Dec  4, 1997

ABSTRACT: Wrapper for L_Reverse

*****************************************************************************/

#define l_reverse L_Reverse

/************************ Function Prototypes *******************************/

extern List *L_Create();
extern int L_Insert(list *l, void *item, int position);
extern void *L_Remove(list *l, int position);
extern int L_Move(list *l, int dir, int no_of_items);
extern int L_Search(list *l, void *item, int (* cmpFnc) (void *a, void *b));
extern int L_Search_N_Remove(list *l, void *item, int (* cmpFnc) (void *a, void*b));
extern int L_LookFor(list *l, void *item, int (* cmpFnc) (void *a, void *b));
extern void *L_Get(list *l, int position);
extern void L_Destroy(list *l, void (*freeFnc) (void *a));
extern List *L_Copy(list *l);
extern void L_Sort(list *l, int (* cmpFnc) (void *a, void *b));
extern void L_Merge(list *l1, list *l2, int (* cmpFnc) (void *a, void *b));
extern void L_Concat(list *l1, list *l2);
extern void L_Randomize(list *l);
extern void L_Reverse(list *l);
extern List *L_Intersect(list *l1, list *l2, int (* cmpFnc) (void *a, void *b));

/* Internal Functions */

extern void L_Error(char *msg);

#endif
