/* 

Copyright 2006: Irmtraud M. Meyer (irmtraud.meyer@cantab.net) and Istvan Miklos 

Some of these functions are based on functions of the HXMATCH software package
by Christina Witwer, see http://www.tbi.univie.ac.at/papers/SUPPLEMENTS/HXMATCH/ 
for more information and a link to the HXMATCH software. Most, but the basic
functions, have been essentially completely rewritten.

This file is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation.
See http://www.gnu.org/copyleft/gpl.html for more information.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include <stdlib.h>

#include "global.h"
#include "process.h"
#include "xutils.h"

void GetOriginalWeightofRefBP(const int iRefN, const INFO *infRefBP, int* iN, INFO** infBP) {

    /* Original weight (owt) of infRefBP is assigned to entries in array infBP */

    int i, j, iFound;

    for (i=0; i<(*iN); i++) { 

	iFound = 0;
	for (j=0; j<iRefN; j++) { 
	    if (((*infBP)[i].bi == infRefBP[j].bi) && ((*infBP)[i].bj == infRefBP[j].bj)) {

		(*infBP)[i].owt = infRefBP[j].owt;
		iFound++;
		break;
	    }
	}
	if (iFound == 0) {
	    printf("ERROR: GetOriginalWeightofRefBP: cannot find reference pair for %d-%d\n",
		   (*infBP)[i].bi, (*infBP)[i].bj);
	}
    }
    return;
}

void CopyBpairs(const int iRefN, const INFO *infRefBP, int* iCopyN, INFO** infCopyBP) {

    /* NOTE: memory of infCopyBP is deleted inside this function and new allocated, 
       it has to be free-ed outside this function */

    if ((iRefN > 0) && (infRefBP != NULL)) {

	(*iCopyN) = iRefN;

	if((*infCopyBP)) free(*infCopyBP);
	(*infCopyBP) = NULL;

	if (!((*infCopyBP) = (INFO *)calloc(iE, sizeof(INFO)))) {
	    fatal("ERROR: CopyBpairs: allocation of INFO* (*infCopyBP) failed.");
	}
	else {
	    int i;
	    for (i=0; i<iRefN; i++) {
		(*infCopyBP)[i].bi  = infRefBP[i].bi;
		(*infCopyBP)[i].bj  = infRefBP[i].bj;
		(*infCopyBP)[i].wt  = infRefBP[i].wt;
		(*infCopyBP)[i].owt = infRefBP[i].owt;
		(*infCopyBP)[i].xx  = infRefBP[i].xx;
		(*infCopyBP)[i].hx  = infRefBP[i].hx;
		(*infCopyBP)[i].col = infRefBP[i].col;
	    }
	}
    }
    return;
}

INFO* InitializeINFO(INFO *inf, const int f, const int p) { // old cset_info

  int i, l;
  
  for (i=f; i<p; i++) {
      inf[i].bi = 0;
      inf[i].bj = 0;
      inf[i].wt = 0;
      inf[i].owt = 0;
      inf[i].xx = 0;
  }
  return inf;
}

int cmp_bi(const void *i1, const void *i2) {
    return ((INFO *)i1)->bi - ((INFO *)i2)->bi;
}

int cmp_bj(const void *j1, const void *j2) {
    return ((INFO *)j1)->bj - ((INFO *)j2)->bj;
}

int cmp_xx(const void *t1, const void *t2) {
    return ((INFO *)t1)->xx - ((INFO *)t2)->xx;
}

int cmp_fct2(const void *t1, const void *t2) {
    if(!((INFO *)t1)->xx)
	return 0;
    return  ((INFO *)t1)->bi - ((INFO *)t2)->bi;
}

int cmp_stwt(const void *sw1, const void *sw2) {
  return ((STACK *)sw2)->wt - ((STACK *)sw1)->wt;
}

int cmp_stbi(const void *si1, const void *si2) {
  return ((STACK *)si1)->bi - ((STACK *)si2)->bi;
}

void ModifyBPWeights(void) { 

    int i;

    GenerateHelices();          /* get helices of iLMin minimum length */
    SelectBpairsInHelices();    /* select bpairs in helices, set bp weight = weight of helix */
    DeleteHelicesAndIndices();
    
    /* sort bpairs */

    qsort(infBpairs, iE, sizeof(INFO), cmp_bj);
    qsort(infBpairs, iE, sizeof(INFO), cmp_bi);

#if _DEBUG
    for (i=0; i<iE; i++) {
	printf("after ModifyBPWeights %d %d %d\n", 
	       infBpairs[i].bi, infBpairs[i].bj, infBpairs[i].wt); 
    }
#endif  
    return;
}

void GenerateHelices(void) {

    // calculates all helices of minimum length iLMin and stores them in global variables
    //
    // array HELIX* helList
    // int iH
    //
    // NOTE: * memory for global arrays iIndx, infPindx and helList have to be freed outside this function
    //       * for iCons >= 0, bpairs (i,i) are not allowed


    const int iCons = 0; // default: 2 * iLMin + LOOP_MIN - 1;
    int i, l, ii, ij, iL;

    // **********************************************************************

    iL_helList = iL_inc;
    iH = 0; // total # of helices

    // ----------------------------------------
    // sort bpairs
    // ----------------------------------------

    qsort(infBpairs, iE, sizeof(INFO), cmp_bj);
    qsort(infBpairs, iE, sizeof(INFO), cmp_bi);
    
    // ----------------------------------------
    // allocate and initialise memory for arrays indx, Pindex and aln_hx
    // ----------------------------------------
    
    if (helList) free(helList);
    helList = NULL;
    if(!(helList = (HELIX *)calloc(iL_helList, sizeof(HELIX))))
	fatal("ERROR: GenerateHelices: allocation of HELIX* helList failed.");

    if (iIndx) free(iIndx);
    iIndx = NULL;
    if (!(iIndx =  (int *) calloc((iV+1), sizeof(int))))
	fatal("ERROR: GenerateHelices: allocation of array int* iIndx failed.");
    for (i = 1; i <= iV; i++) // Irmi: what about element [0] ?!
	iIndx[i] = (i*(i-1))/2;
    
    if (infPindx) free(infPindx);
    infPindx = NULL;
    iL = (iV * (iV+1) ) / 2;
    if (!(infPindx = (INFO **) calloc(iL, sizeof(INFO *))))
	fatal("ERROR: GenerateHelices: allocation of array INFO** infPindx failed.");
    
    if (helList) free(helList);
    helList = NULL;
    if (!(helList = (HELIX *)calloc(iL_helList, sizeof(HELIX)))) 
	fatal("ERROR: GenerateHelices: allocation of array HELIX* helList failed.");

    for(i=0; i<iE; i++) {
	ii = infBpairs[i].bi;
	ij = infBpairs[i].bj;
	infPindx[iIndx[ij]+ii] = &infBpairs[i];
#ifdef _DEBUG
	fprintf(stdout, "bp %d-%d [%d] infPindx %d\n", ii, ij, i, infPindx[iIndx[ij]+ii]);
#endif
    }

     // **********************************************************************
    
    int iIndex, iHLength, iHWeight;

    for (ii=1; ii <= iV-iCons+1; ii++) {
	for(ij=iV; ij >= ii+iCons+1; ij--) {
	    
#ifdef _DEBUG
	    fprintf(stdout, "%d-%d ?\n", ii, ij);
#endif
	    if (! infPindx[iIndx[ij]+ii]) continue;
	    else { // if ii-ij bpair

		iIndex = (int) (infPindx[iIndx[ij]+ii]-infBpairs);

		if (infBpairs[iIndex].wt > 0) { // if ii-ij has wt > 0
#ifdef _DEBUG
		    fprintf(stdout, "%d-%d i %d wt %d Pindx %d\n",
			    ii, ij, iIndex, infBpairs[iIndex].wt, infPindx[iIndx[ij]+ii]);
#endif
		    // skip bp ii-ij if outer bp exists
		    
		    if (ii > 1 && ij < iV) {
			if (infPindx[iIndx[ij+1]+ii-1]) {
#ifdef _DEBUG
			    fprintf(stdout, "outer bp exists %d-%d Pindx %d => skip bp %d-%d\n",
				    ii-1, ij+1, infPindx[iIndx[ij+1]+ii-1], ii, ij);
#endif
			    continue;
			}
		    }
		    iHLength = 0;
		    iHWeight = 0;

		    GetHelix(ii, ij, &iHLength, &iHWeight);
#ifdef _DEBUG
		    fprintf(stdout, "%d-%d: iHLength %d iHWeight %d \n", ii, ij, iHLength, iHWeight);
#endif
		    if (iHLength >= iLMin) { // if helix has minimum length 

			// allocate and initialise more memory for helices, if necessary

			if (iH >= iL_helList) {
			    if (!(helList = (HELIX *)realloc(helList, (iL_helList+iL_inc)*sizeof(HELIX))))
				fatal("HelForm(): realloc of helList failed.");
			    for(l=iL_helList; l<iL_helList+ iL_inc; l++) {
				helList[l].iL     =0;
				helList[l].bp[0]  =0;
				helList[l].bp[1]  =0;
				helList[l].iW     =0;
			    }
			    iL_helList += iL_inc;
			}
			// save information on new helix in helList: length, outer base-pair

			helList[iH].iL    = iHLength;
			helList[iH].bp[0] = ii;
			helList[iH].bp[1] = ij;
			helList[iH].iW    = iHWeight;
#ifdef _DEBUG
			fprintf(stdout, "helList[%d] iL %d iW %d %d-%d\n", 
				iH, helList[iH].iL, helList[iH].iW, helList[iH].bp[0], helList[iH].bp[1]);
#endif
			iH++;
		    }
		}
	    }
	}
    }
    return;
}

void GetHelix(const int ii, const int ij, int* iHL, int* iHW) {

    // uses global arrays infBpairs, infPindx and iIndx and calculates
    // length (iHL) and weight (iHW) of helix starting at ii-ij

    int iIndex, ji, jj;
    
    iIndex = (int) (infPindx[iIndx[ij]+ii]-infBpairs);

    (*iHL) = 1;
    (*iHW) = infBpairs[iIndex].wt;

#ifdef _DEBUG
    fprintf(stdout, "* GetHelix: %d-%d [%d] iHW + %d => %d iHL %d\n", 
	    ii, ij, iIndex, infBpairs[iIndex].wt, *iHW, *iHL);
#endif
    for(;;) { // loop over inner bps 

	ji = ii+(*iHL);
	jj = ij-(*iHL);

	// if minloop condition is violated
	if (ij - ii - 2 * (*iHL) <= LOOP_MIN) break;
	// if end of stack is reached
	if (! infPindx[iIndx[jj]+ji]) break;

	iIndex = (int) (infPindx[iIndx[jj]+ji]-infBpairs);
	(*iHL)++;
	(*iHW) += infBpairs[iIndex].wt;
#ifdef _DEBUG
	fprintf(stdout, "* GetHelix: %d-%d [%d] iHW + %d => %d iHL %d\n", 
		ji, jj, iIndex, infBpairs[iIndex].wt, *iHW, *iHL);
#endif
    }
    return;
}

void SelectBpairsInHelices(void) {

    // selects all bpairs in global array INFO* infBpairs that are contained in 
    // global array HELIX* helList and assigns weight of helix to bpairs in helix

    int i, j, k, n, iL, bi, bj, iW;
    INFO* infTempBpairs= NULL; // array of selected bpairs
 
    if(!(infTempBpairs = (INFO *)calloc(iE+1, sizeof(INFO)))) {
	fatal("ERROR: SelectBpairsInHelices: allocation of infTempBpairs failed.");
    }
  
    j = 0;  

    for (i=0; i<iH; i++) { // loop over helices

	iL = helList[i].iL;
	iW = helList[i].iW;

	for (k=0; k<iL; k++) {

	    bi = helList[i].bp[0] + k;
	    bj = helList[i].bp[1] - k;

	    infTempBpairs[j].bi = bi;
	    infTempBpairs[j].bj = bj;
	    infTempBpairs[j].wt = iW;
#ifdef _DEBUG
	    fprintf(stdout, "selected %d. bp in helix %d: %d-%d wt %d\n", 
		    k+1, i, infTempBpairs[j].bi, infTempBpairs[j].bj, infTempBpairs[j].wt);
#endif
	    j++;
	}
    }
    if (infBpairs) free(infBpairs);

    // save selected bpairs in global variables INFO* infBpairs and int iE

    infBpairs = infTempBpairs;
    iE = j;

    return;  
}

void DeleteHelicesAndIndices(void) {

    if (iIndx) free(iIndx);
    iIndx = NULL;
    if (infPindx) free(infPindx);
    infPindx = NULL;
    if (helList) free(helList);
    helList = NULL;
    iH = 0;

    return;
}

void mwm2bpairs(const int* iMate, INFO** infNewBpairs, int* iNewE) {

    /* 
       mwm2bpairs: 

       input:  int* iMate, int iV (global), int iE (global), INFO* infBpairs (global)
       output: INFO* infNewBpairs, int iNewE

       function converts bpairs in array iMate into array INFO* infNewBpairs, 
       whose number of entries is iNewE
    */

    int k, i, j;

    (*iNewE) = 0;               /* iNewE = 2 * (total # of edges in MWM matching) */
    for (i = 1; i <= iV; i++) { /* derive iNewE */
	if(iMate[i] != 0)
	    (*iNewE)++;
    }
    if((*iNewE)%2)
	fatal("ERROR: mwm2bpairs: total number of entries in array int* iMate is not even.\n");
    (*iNewE) = (*iNewE)/2; /* iMate contains each bp twice, infOutBpairs only once */

#ifdef _DEBUG
    printf("iNewE = %d\n", (*iNewE));
#endif
    
    if (*infNewBpairs) free(*infNewBpairs);
    *infNewBpairs = NULL;
    if (!((*infNewBpairs) = (INFO *)calloc((*iNewE), sizeof(INFO))))
	fatal("ERROR: mwm2bpairs: allocation of infNewBpairs failed.\n");

    k = 0;

    for (i = 1; i <= iV; i++) {
	if (i < iMate[i]) {
	    for (j = 0; j < iE; j++) {
		if ((infBpairs[j].bi == i) && (infBpairs[j].bj == iMate[i])) {

		    (*infNewBpairs)[k].bi = i;
		    (*infNewBpairs)[k].bj = iMate[i];
		    (*infNewBpairs)[k].xx = 1;
		    (*infNewBpairs)[k].wt = infBpairs[j].wt;
		    (*infNewBpairs)[k].owt = infBpairs[j].owt;
		    k++;
		    break;
		}
	    }
	}
    }
    return;
}

void MarkBpairs(int* iN_BP, INFO** infBP) {

    /*
      MarkBpairs: 

      input: array INFO* infBP with iN_BP entries
      output = input

      NOTE: all bp in infBP are taken into account, but only the status of bps with xx != 0 is changed

      function first sets all bps with xx!=0 to xx=1 and then classifies each bp with xx!=0 as:

      xx = 0 if bp is isolated
      xx = 2 if bp is outer bp of helix
      xx = 3 if bp is inner bp of helix
      xx = 1 if bp is internal bp of helix

    */

    int i, iO, iI, iIsol; 

    for (i = 0; i < (*iN_BP); i++) {
	if ((*infBP)[i].xx) {
	    (*infBP)[i].xx = 1;
	}
    }
    qsort((*infBP), (*iN_BP), sizeof(INFO), cmp_xx);   /* sort entries in (*infBP) by increasing xx values */
    qsort((*infBP), (*iN_BP), sizeof(INFO), cmp_fct2); /* sort entries in (*infBP) by increasing bi values */

    /* 
       iO marks the outer bp of a helix (xx = 2) 
       iI marks the inner bp of a helix (xx = 3)
       iIsol = total # of isolated base-pairs 
    */  

    iIsol = 0; 

    /* deal with first bp [0] */

    if ((*infBP)[0].xx != 0) {
	if (! (((*infBP)[1].bi == (*infBP)[0].bi+1) &&  /* if not 0(1(..1)0) [if first bp isolated] */
	       ((*infBP)[1].bj == (*infBP)[0].bj-1))) {
	    (*infBP)[0].xx = 0;
	    iIsol++;
	}
	else (*infBP)[0].xx = 2; /* mark bp 0 as outer bp of stack */
    }
    /* deal with last bp [(*iN_BP)-1] */

    if ((*infBP)[(*iN_BP)-1].xx != 0) {
	if (!(((*infBP)[(*iN_BP)-2].bi == (*infBP)[(*iN_BP)-1].bi-1) && /* if not k-2(k-1(..k-1)k-2) [if last bp isolated] */
	      ((*infBP)[(*iN_BP)-2].bj == (*infBP)[(*iN_BP)-1].bj+1))) {
	    (*infBP)[(*iN_BP)-1].xx = 0;
	    iIsol++;
	}
	else (*infBP)[(*iN_BP)-1].xx = 3; /* mark bp k-1 as inner bp of stack */
    }

    for (i = 1; i < (*iN_BP)-1; i++) { /* loop over all bpairs except the first and last */
	
	/* determine status of bp i */
	
	if (!((*infBP)[i].xx)) continue; /* skip base-pair if xx == 0 */

	iO = iI = 0;
	if ( !(((*infBP)[i-1].bi == (*infBP)[i].bi-1) &&  /* if not i-1(i(...i)i-1) */
	       ((*infBP)[i-1].bj == (*infBP)[i].bj+1))) {
	    (*infBP)[i].xx = 2; /* mark bp i as outer bp */
	    iO =1;
	}
	if ( !(((*infBP)[i+1].bi == (*infBP)[i].bi+1) && /* if not i(i+1(...i+1)i) */
	       ((*infBP)[i+1].bj == (*infBP)[i].bj-1))) {
	    (*infBP)[i].xx = 3; /* mark bp i as inner bp */
	    iI = 1;
	}
	if (iO && iI) { /* if not i-1(i(...i)i-1) AND not i(i+1(...i+1)i) */
	    (*infBP)[i].xx = 0; /* mark bp i as isolated */
	    iIsol++;
	}
    }
    qsort((*infBP), (*iN_BP), sizeof(INFO), cmp_xx); /* sort entries in array INFO* (*infBP) in asc. order of xx */

    return;
}

void ExtendStacks(const int iRefN, const INFO *infRefBpairs, int* iN, INFO** infBpairs)  {

    /* 

    ExtendStacks:
    
    input: arguments above
    output: int iN, INFO* infBpairs

    function takes iRefN bps in reference list infRefBpairs and checks if any of the bps in 
    list infBpairs can be further extended (xx values are changed accordingly)

    NOTE: the xx-values of infBpairs have to properly indicate whether a bp is isolated (xx=0), internal (xx=1)
          or at the inside (xx=3) or outside (xx=2) of a stack 

    */

    // extend_stacks: take array INFO* bpairs and modify array INFO* outcom by
    // extending the helices by good base-pairs in infRefBpairs, if possible

    int bi, bj, iExtend;
    int i, j, k, iL;

    iL = (*iN);

    for (i = 0; i < (*iN); i++) {
	switch ((*infBpairs)[i].xx) {

	    case 0: /* if bp isolated => skip */
		break;
	    case 1: /* if bp inner bp => skip */
		break;
	    case 2: /* if bp i outer bp */
		
		/* check bp i can be extended with by new bp bi-bj outside */

		bi = (*infBpairs)[i].bi - 1;
		bj = (*infBpairs)[i].bj + 1;
		iExtend = 1;
		
		for (k = 0; k < iRefN; k++) {
		    if ((bi == infRefBpairs[k].bi) && (bj == infRefBpairs[k].bj)) { /* if bi-bj in ref list */
			
			/* skip bi-bj if one base is already contained in (*infBpairs) with xx != 0 */
			for (j = 0; j < (*iN); j++) {
			    if (((*infBpairs)[j].xx) &&
				((bi == (*infBpairs)[j].bi) || (bi == (*infBpairs)[j].bj) ||
				 (bj == (*infBpairs)[j].bi) || (bj == (*infBpairs)[j].bj) )) {
				iExtend = 0; /* don't take the new bp ... */
				break;       /* ... go to next k-value */
			    }
			}
			if (iExtend) { /* if bi-bj not in ref list => add to list infBpairs */
			    /* allocate more memory for (*infBpairs), if required */
			    if ((*iN) >= iL) {
				if (!((*infBpairs) = (INFO *)realloc((*infBpairs), (iL+INC)*sizeof(INFO))))
				    fatal("ERROR: ExtendStacks: realloc of (*infBpairs) failed.");
				InitializeINFO((*infBpairs), (*iN), (*iN) +INC);	     
				iL += INC;
			    }
			    /* add new bp bi-bj to list (*infBpairs) */
			    (*infBpairs)[(*iN)].bi = bi;
			    (*infBpairs)[(*iN)].bj = bj;
			    (*infBpairs)[(*iN)].xx = (*infBpairs)[i].xx;   /* new bp becomes outside bp of stack */
			    (*infBpairs)[(*iN)].owt = infRefBpairs[k].owt;
			    (*infBpairs)[(*iN)++].wt = infRefBpairs[k].wt;

			    (*infBpairs)[i].xx = 1;                         /* old bp i is now inside bp */
			}
		    }
		}
		break;
	    case 3: /* if bp inner bp */

		bi = (*infBpairs)[i].bi + 1;
		bj = (*infBpairs)[i].bj - 1;
		iExtend = 1;

		for (k = 0; k < iRefN; k++) {
		    if ((bi == infRefBpairs[k].bi) && (bj == infRefBpairs[k].bj)) { /* if bi-bj in ref list */

			/* skip bi-bj if one base is already contained in (*infBpairs) with xx != 0 */
			for (j = 0; j < (*iN); j++) {
			    if (((*infBpairs)[j].xx) &&
				((bi == (*infBpairs)[j].bi) || (bi == (*infBpairs)[j].bj) ||
				 (bj == (*infBpairs)[j].bi) || (bj == (*infBpairs)[j].bj) )) {
				iExtend = 0; /* don't take the new bp ... */
				break;       /* ... go to next k-value    */
			    }
			}
			if (iExtend) {  /* if bi-bj not in ref list => add to list infBpairs */
			    /* allocate more memory for (*infBpairs), if required */
			    if ((*iN) >= iL) {
				if (!((*infBpairs) = (INFO *)realloc((*infBpairs), (iL+INC)*sizeof(INFO))))
				    fatal("ERROR: ExtendStacks: realloc of (*infBpairs) failed.");	   
				InitializeINFO((*infBpairs), (*iN), (*iN)+INC);
				iL += INC;
			    }
			    /* add new bp bi-bj to list (*infBpairs) */
			    (*infBpairs)[(*iN)].bi = bi;
			    (*infBpairs)[(*iN)].bj = bj;
			    (*infBpairs)[(*iN)].xx = (*infBpairs)[i].xx;    /* new bp becomes inside bp of stack */
			    (*infBpairs)[(*iN)].owt = infRefBpairs[k].owt;
			    (*infBpairs)[(*iN)++].wt = infRefBpairs[k].wt;
			    
			    (*infBpairs)[i].xx = 1;                         /* old bp i becomes internal bp */
			}
		    }
		}
		break;
	    default:
		fprintf(stderr, "ERROR: ExtendStacks: unknown tag %d.\n", (*infBpairs)[i].xx);
	}
    }
    return;
}

int PostProcess(const int iN, const int* iMate, int* iNE, INFO** infBP, char** cStrucString) { 

    int iStrucChanged = 0;
    int i;
 
    int iNewE;
    INFO* infMWMBpairs = NULL;

    int iS = 0;
    STACK* stStacks = NULL;

    mwm2bpairs(iMate, &infMWMBpairs, &iNewE);

#ifdef _DEBUG
    for (i=0; i<iNewE; i++) { 
	printf("mwm2bpairs %d-%d wt %d xx %d\n", 
	       infMWMBpairs[i].bi, infMWMBpairs[i].bj, infMWMBpairs[i].wt, infMWMBpairs[i].xx);
    }
#endif

    MarkBpairs(&iNewE, &infMWMBpairs);

#ifdef _DEBUG
    for (i=0; i<iNewE; i++) { 
	printf("MarkBpairs %d-%d wt %d xx %d\n", 
	       infMWMBpairs[i].bi, infMWMBpairs[i].bj, infMWMBpairs[i].wt, infMWMBpairs[i].xx);
    }
#endif

    ExtendStacks((*iNE), (*infBP), &iNewE, &infMWMBpairs);

#ifdef _DEBUG
    for (i=0; i<iNewE; i++) { 
	printf("ExtendStacks %d-%d wt %d xx %d\n", 
	       infMWMBpairs[i].bi, infMWMBpairs[i].bj, infMWMBpairs[i].wt, infMWMBpairs[i].xx);
    }
#endif

    qsort(infMWMBpairs, iNewE, sizeof(INFO), cmp_fct2);

#ifdef _DEBUG
    for (i=0; i<iNewE; i++) { 
	printf("cmp_fct2 %d-%d wt %d xx %d\n", 

	       infMWMBpairs[i].bi, infMWMBpairs[i].bj, infMWMBpairs[i].wt, infMWMBpairs[i].xx);
    }
#endif
    
    GetStacks(&iNewE, &infMWMBpairs, &iS, &stStacks);

#ifdef _DEBUG    
    for (i=0; i<iNewE; i++) { 
	printf("GetStacks %d-%d wt %d xx %d hx %d\n", 
	       infMWMBpairs[i].bi, infMWMBpairs[i].bj, infMWMBpairs[i].wt, 
	       infMWMBpairs[i].xx, infMWMBpairs[i].hx);
    }
    for (i=0; i<iS; i++) { 
	printf("GetStacks bi %d bj %d len %d wt %d col %d \n", 
	       stStacks[i].bi, stStacks[i].bj, stStacks[i].len, 
	       stStacks[i].wt, stStacks[i].col);
    }
#endif
    ColorStacks(&iS, &stStacks);

#ifdef _DEBUG
    for (i=0; i<iS; i++) { 
	printf("ColorStacks bi %d bj %d len %d wt %d col %d \n", 
	       stStacks[i].bi, stStacks[i].bj, stStacks[i].len, 
	       stStacks[i].wt, stStacks[i].col);
    }
#endif
    ColorBpairs(iS, stStacks, &iNewE, &infMWMBpairs);

#ifdef _DEBUG
    for (i=0; i<iNewE; i++) { 
	printf("ColorBpairs %d-%d wt %d xx %d col %d\n", 
	       infMWMBpairs[i].bi, infMWMBpairs[i].bj, infMWMBpairs[i].wt, 
	       infMWMBpairs[i].xx, infMWMBpairs[i].col);
    }
#endif

    ExtendStacks((*iNE), (*infBP), &iNewE, &infMWMBpairs);

#ifdef _DEBUG
    for (i=0; i<iNewE; i++) { 
	printf("ExtendStacks %d-%d wt %d xx %d\n", 
	       infMWMBpairs[i].bi, infMWMBpairs[i].bj, infMWMBpairs[i].wt, infMWMBpairs[i].xx);
    }
#endif

    (*cStrucString) = PrintStructure(iV, iNewE, infMWMBpairs, (*cStrucString), &iStrucChanged);

#ifdef _DEBUG
    printf("cStrucString %s\n", (*cStrucString));
#endif

    GetNewBpairs(iNewE, infMWMBpairs, &iE, &infBpairs);

#ifdef _DEBUG
    for (i=0; i<iE; i++) { 
	printf("GetNewBpairs %d-%d wt %d owt %d xx %d\n", 
	       infBpairs[i].bi, infBpairs[i].bj, infBpairs[i].wt, infBpairs[i].owt, infBpairs[i].xx);
    }
#endif

    if (stStacks) free(stStacks);
    stStacks = NULL;

    if (infMWMBpairs) free(infMWMBpairs);
    infMWMBpairs = NULL;

    return(iStrucChanged);
}

void GetStacks(int* iNBP, INFO** infBP, int* iN, STACK** stStacks) {

    // stacks: allocates and initialises memory for array STACK* stack,
    //         every helix (=stack) corresponds to one entry in array stack,
    //         total weight, length and the coordinates of the starting bp
    //         are saved for each helix
    
    int i, k, iL, iW;
    
    /* count iN = # of stacks */

    (*iN) = 0;
    for(i = 0; i < (*iNBP); i++) {
	if ((*infBP)[i].xx == 2)
	    (*iN)++;
    }

    if ((*stStacks)) free((*stStacks));
    *stStacks = NULL;
    if (!((*stStacks) = (STACK *)calloc((*iN), sizeof(STACK))))
	fatal("ERROR: GetStacks: allocation of stack failed!!!");

    k = 0;

    for (i = 0; i < (*iNBP); i++) {

	iL = 0;   // length of helix 
	iW = 0;   // weight of helix
	
	if ((*infBP)[i].xx == 2) { /* if outside bp of helix */

	    (*infBP)[i].hx = k;                   /* helix number to bp i */
	    (*stStacks)[k].bi = (*infBP)[i].bi;   /* left base of bp outside of helix k */
	    (*stStacks)[k].bj = (*infBP)[i].bj;   /* right base of bp outside of helix k */

	    iL++;                            
	    iW += (*infBP)[i].wt;

	    /* loop over internal bps of helix */

	    while ((*infBP)[i+iL].xx == 1) {
		(*infBP)[i+iL].hx = k;
		iW += (*infBP)[i+iL].wt;
		iL++;
	    }

	    /* end of helix */ 

	    (*infBP)[i+iL].hx = k;
	    iW += (*infBP)[i+iL].wt;
	    iL++;

	    (*stStacks)[k].len = iL;  /* length of helix k */
	    (*stStacks)[k].wt = iW;   /* weight of helix k */
	    k++;               
	}
    }
    return;
}

void ColorStacks(const int* iS, STACK** stStacks) { 

    /*  NOTE: coloring starts with 1 */

    int i, j, ii, ij, ji, jj, icol, jcol, iNextCol, iMoreCol;

    /* order entries in stack by decreasing weight */

    qsort((*stStacks), (*iS), sizeof(STACK), cmp_stwt); 

#ifdef _DEBUG
    for (i=0; i<(*iS); i++) { 
	printf("cmp_stwt [%d] %d-%d len %d wt %d col %d \n", 
	       i, (*stStacks)[i].bi, (*stStacks)[i].bj, (*stStacks)[i].len, 
	       (*stStacks)[i].wt, (*stStacks)[i].col);
    }
#endif
    /* color stacks:
       
    strategy: - go through list of stacks with decreasing weight
                if stacks is incompatible with any stack of heigher weight => assign it next color
    */

    iNextCol = 0;
    iMoreCol = 1;

    /* at the start, color all stacks with 0 */

    for (i=0; i<(*iS); i++) {
	(*stStacks)[i].col = 0;
    }

    /* while less than 3 colors have been assigned and we can assign new colors */

    while ((iNextCol <= 2) && (iMoreCol > 0)) { 

	iMoreCol = 0;
	iNextCol++;

#ifdef _DEBUG
	printf("iMoreCol %d iNextCol %d\n", iMoreCol, iNextCol);
#endif
	for (i=0; i<(*iS); i++) {

	    ii = (*stStacks)[i].bi;
	    ij = (*stStacks)[i].bj;
	    icol = (*stStacks)[i].col;
#ifdef _DEBUG
	    printf("\t [%d] col %d\n", i, icol);
#endif
	    if (icol == (iNextCol - 1)) { /* consider only stacks of the last color */

		for (j=0; j<i; j++) {

		    ji = (*stStacks)[j].bi;
		    jj = (*stStacks)[j].bj;
		    jcol = (*stStacks)[j].col;
#ifdef _DEBUG		    
		    printf("\t\t [%d] col %d\n", j, jcol);
#endif
		    if (icol == jcol) { /* if stack i and j have the same color */
			if (! ((ij < ji) || (jj < ii))) { /* if stack i and j are incompatible */
			
			    iMoreCol++;
			    (*stStacks)[i].col = iNextCol; /* assign next color to stack i */
#ifdef _DEBUG
			    printf("\t\t [%d] col %d not compatible with [%d] col %d\n", 
				   j, jcol, i, (*stStacks)[i].col);
#endif
			    break;                            /* go to next stack i */
			}
		    }
		}
	    }
	}
    }

    // increase all colours by 1

    for (i=0; i<(*iS); i++) {
	(*stStacks)[i].col++;
    }
    
    return;
}

void ColorBpairs(const int iS, const STACK* stStacks, const int* iNBP, INFO** infBP) {

    int i, k;

    /* order stacks by increasing bi coordinate */    

    qsort(stStacks, iS, sizeof(STACK), cmp_stbi); 

    /* color bps according to color of their stack */

    for (k=0; k<iS; k++) {

#ifdef _DEBUG
	printf("inside ColorBpairs dealing with bps of stack %d %d-%d wt %d len %d col %d\n", 
	       k, stStacks[k].bi, stStacks[k].bj, stStacks[k].wt, stStacks[k].len, stStacks[k].col);
#endif
	for (i=0; i<(*iNBP); i++) {

	    /* if bp i starts stack k */
	    if ((stStacks[k].bi == (*infBP)[i].bi) &&
		(stStacks[k].bj == (*infBP)[i].bj)) {

		/* assign all base-pairs of that stack the same color */
		
		(*infBP)[i].col = stStacks[k].col;
#ifdef _DEBUG
		printf("inside ColorBpairs %d-%d wt %d xx %d col %d\n", 
		       (*infBP)[i].bi, (*infBP)[i].bj, (*infBP)[i].wt, 
		       (*infBP)[i].xx, (*infBP)[i].col);
#endif
		i++;
		while ((*infBP)[i].xx != 2) {
		    (*infBP)[i].col = stStacks[k].col;
#ifdef _DEBUG
		    printf("inside loop of ColorBpairs %d-%d wt %d xx %d col %d\n", 
			   (*infBP)[i].bi, (*infBP)[i].bj, (*infBP)[i].wt, 
			   (*infBP)[i].xx, (*infBP)[i].col);
#endif
		    i++;
		    if (i == (*iNBP)) break;
		}
	    }
	}
    }  
    return;
}

void GetNewBpairs(const int iNBP, const INFO *infBP, int* iModNBP, INFO** infModBP) {

    /* takes reference list of bps (infBP) and retains only those bps infModBP that
       are contained in infBP and have status xx != 0 */

    INFO* infTempBP = NULL; /* temporary list of bps */
    int iNewBP;

    int i, j, rbp, ct;
    int ii, ij, ji, jj;
    int conflict0, conflict1;

    /* count retained bps */

    iNewBP = 0;
    for (i=0; i<iNBP; i++) {
	if (infBP[i].xx && ((infBP[i].col == 1) || (infBP[i].col == 2))) 
	    iNewBP++;
    }

    /* allocate memory */
    
    if (!(infTempBP = (INFO *)calloc(iNewBP+1, sizeof(INFO)))) {
	fatal("GetNewBpairs: memory allocation failed.");
    }

    iNewBP = 0;
    for (i=0; i<iNBP; i++) {
	if (infBP[i].xx && ((infBP[i].col == 1) || (infBP[i].col == 2))) {

	    infTempBP[iNewBP].bi = infBP[i].bi;
	    infTempBP[iNewBP].bj = infBP[i].bj;
	    infTempBP[iNewBP].wt = infBP[i].wt;
	    infTempBP[iNewBP].owt = infBP[i].owt;
	    infTempBP[iNewBP].xx = infBP[i].xx;
	    infTempBP[iNewBP].hx = 0;
	    infTempBP[iNewBP].col = 0;

	    iNewBP++;
	}
    }
    
    qsort(infTempBP, iNewBP, sizeof(INFO), cmp_bj); // sort bpairs by increasing bj values
    qsort(infTempBP, iNewBP, sizeof(INFO), cmp_bi); // sort bpairs by increasing bi values
    
    /* free memory of old list of bps */

    if ((*infModBP)) free(*infModBP);

    /* assign modified values to output parameters */

    (*infModBP) = infTempBP;
    (*iModNBP) = iNewBP;

    return;
}
